#include <SoftSerial_INT0.h>  // https://github.com/J-Rios/Digispark_SoftSerial-INT0
#include "DigiKeyboard.h"

// Some additional Key Definitions missing in the DigiSpark Headers
#define KEY_ESCAPE 0x29
#define KEY_BACKSPACE 0x2A
#define KEY_TAB 0x2B
#define KEY_DELETE 0x4C

#define KEY_KEYPAD_0 0x62
#define KEY_KEYPAD_1 0x59
#define KEY_KEYPAD_2 0x5A
#define KEY_KEYPAD_3 0x5B
#define KEY_KEYPAD_4 0x5C
#define KEY_KEYPAD_5 0x5D
#define KEY_KEYPAD_6 0x5E
#define KEY_KEYPAD_7 0x5F
#define KEY_KEYPAD_8 0x60
#define KEY_KEYPAD_9 0x61

#define KEY_F1 0x3A
#define KEY_F2 0x3B
#define KEY_F3 0x3C
#define KEY_F4 0x3D
#define KEY_F5 0x3E
#define KEY_F6 0x3F
#define KEY_F7 0x40
#define KEY_F8 0x41
#define KEY_F9 0x42
#define KEY_F10 0x43
#define KEY_F11 0x44
#define KEY_F12 0x45

#define KEY_INSERT 0x49
#define KEY_HOME 0x4A
#define KEY_PAGEUP 0x4B
#define KEY_DEL 0x4C
#define KEY_END 0x4D
#define KEY_PAGEDOWN 0x4E
#define KEY_ARROW_RIGHT 0x4F
#define KEY_ARROW_LEFT 0x50
#define KEY_ARROW_DOWN 0x51
#define KEY_ARROW_UP 0x52

#define KEY_LEFT_CTRL 0xE0
#define KEY_LEFT_SHIFT 0xE1
#define KEY_LEFT_ALT 0xE2
#define KEY_LEFT_GUI 0xE3

const unsigned int KEYPAD[] PROGMEM = { KEY_KEYPAD_0, KEY_KEYPAD_1, KEY_KEYPAD_2, KEY_KEYPAD_3, KEY_KEYPAD_4, KEY_KEYPAD_5, KEY_KEYPAD_6, KEY_KEYPAD_7, KEY_KEYPAD_8, KEY_KEYPAD_9 };
const char TRANSLATE_KEYPAD[] PROGMEM = {'!', '"', '#', '$', '%', '&', 39, '(', ')', '*',
                                         '+', ',', '-', '.', '/', ':', ';', '<', '=', '>',
                                         '?', '@', 'A', 'Q', 'Y', 'Z', '[', ']', '\\', '^',
                                         '_', '`', 'a', 'q', 'y', 'z', '{', '|', '}', '~',
                                        };

const char TRANSLATE_CONTR[][2] = { { '\n', KEY_ENTER }, { '\r', KEY_ENTER }, { '\e', KEY_ESCAPE }, { 127, KEY_BACKSPACE }, { 8, KEY_BACKSPACE }, { 9, KEY_TAB} };

const char TRANSLATE_FUNC[][5] = { { 79, 80, 0, 0, KEY_F1 }, { 91, 49, 49, 126, KEY_F1 }, { 79, 81, 0, 0, KEY_F2 }, { 91, 49, 50, 126, KEY_F2 },
  { 79, 82, 0, 0, KEY_F3 }, { 91, 49, 51, 126, KEY_F3 }, { 79, 84, 0, 0, KEY_F4 }, { 91, 49, 52, 126, KEY_F4 },
  { 91, 49, 54, 126, KEY_F5 }, { 91, 49, 55, 126, KEY_F6 }, { 91, 49, 56, 126, KEY_F7 }, { 91, 49, 57, 126, KEY_F8 },
  { 91, 50, 48, 126, KEY_F9 }, { 91, 50, 49, 126, KEY_F10 }, { 91, 50, 51, 126, KEY_F11 }, { 91, 50, 52, 126, KEY_F12 },
  { 91, 50, 126, 0, KEY_INSERT }, { 91, 51, 126, 0, KEY_DEL }, { 91, 49, 126, 0, KEY_HOME }, { 91, 52, 126, 0, KEY_END },
  { 91, 53, 126, 0, KEY_PAGEUP }, { 91, 54, 126, 0, KEY_PAGEDOWN }, { 91, 65, 0, 0, KEY_ARROW_UP }, { 91, 66, 0, 0, KEY_ARROW_DOWN },
  { 91, 67, 0, 0, KEY_ARROW_RIGHT }, { 91, 68, 0, 0, KEY_ARROW_LEFT }
};

SoftSerial mySerial(2, 1); // RX on 2 (fixed) and TX on 1

// Translate given Character into Keyboard Input
void KeyboardWriteTranslate(char c) {
  unsigned int i;
  int found = 0;
  // Search if character is in first translation table
  for (i = 0; i < sizeof(TRANSLATE_KEYPAD); i++) {
    if (c == pgm_read_byte(TRANSLATE_KEYPAD + i)) {
      found = 1;
      break;
    }
  }
  if (found == 1) { // If character is in the first table (a blacklist) then type the character via ALT+Keypad
    if (c >= 100) {
      //      DigiKeyboard.sendKeyStroke(pgm_read_byte(KEYPAD + (c / 100)));
      DigiKeyboard.sendKeyPress(pgm_read_byte(KEYPAD + (c / 100)), MOD_ALT_LEFT);
      DigiKeyboard.sendKeyPress(0, MOD_ALT_LEFT);
    }
    if (c >= 10) {
      //      DigiKeyboard.sendKeyStroke(pgm_read_byte(KEYPAD + ((c / 10) % 10)));
      DigiKeyboard.sendKeyPress(pgm_read_byte(KEYPAD + ((c / 10) % 10)), MOD_ALT_LEFT);
      DigiKeyboard.sendKeyPress(0, MOD_ALT_LEFT);
    }
    //DigiKeyboard.sendKeyStroke(pgm_read_byte(KEYPAD + (c % 10)));
    DigiKeyboard.sendKeyPress(pgm_read_byte(KEYPAD + (c % 10)), MOD_ALT_LEFT);
    DigiKeyboard.sendKeyPress(0, MOD_ALT_LEFT);
    DigiKeyboard.delay(10); // Wait 10ms
    DigiKeyboard.sendKeyPress(0, 0);
  } else {
    found = 0; // Search in second table (ESC, Enter, Backspace, Tab)
    for (i = 0; i < (sizeof(TRANSLATE_CONTR) / 2); i++) {
      if (c == TRANSLATE_CONTR[i][0]) {
        found = 1;
        c = TRANSLATE_CONTR[i][1];
      }
    }
    if (found == 1) { // And if found send the corresponding keycode
      DigiKeyboard.sendKeyPress(c, 0);
      DigiKeyboard.sendKeyPress(0, 0);
    } else if (c <= 26 && c > 0) { // If 1-26 (CTRL-A to CTRL-Z) then type this way (this is not [aq]wert[yz] aware!)
      DigiKeyboard.sendKeyPress((c + KEY_A) - 1, MOD_CONTROL_LEFT);
      DigiKeyboard.sendKeyPress(0, 0);
    } else { // Normal Character - send via Library function
      DigiKeyboard.write(c);
    }
  }
}

void setup() {
  mySerial.begin(19200);  // 19200 Baud - is more stable than 9600!
}

void loop() {
  byte c;
  int i, j, found = 0;
  static char buffer[5];
  static int n = 0;
  static unsigned long ESCFOUND = 0;
  
  if (ESCFOUND != 0) {
    if (millis() - ESCFOUND > 200) { // If we found a ESC Character but 200ms nothing then just send what we buffered
      DigiKeyboard.sendKeyPress(KEY_ESCAPE, 0);
      DigiKeyboard.sendKeyPress(0, 0);
      for (j = 0; j < n; j++) {
        KeyboardWriteTranslate(buffer[j]);
      }
      ESCFOUND = 0;
      n = 0;
      for (j = 0; j < 5; j++) {
        buffer[j] = '\0';
      }
    }
  }
  
  if (mySerial.available()) {
    c = mySerial.read();
    if (ESCFOUND != 0) { // If ESC was found earlier 
      if (n >= 4) { // Buffer full -> Send out and stop searching
        DigiKeyboard.sendKeyPress(KEY_ESCAPE, 0);
        DigiKeyboard.sendKeyPress(0, 0);
        for (j = 0; j < n; j++) {
          KeyboardWriteTranslate(buffer[j]);
        }
        ESCFOUND = 0;
        n = 0;
        for (j = 0; j < 5; j++) {
          buffer[j] = '\0';
        }
      } else { // Keep next character
        buffer[n] = c;
        n++;
        buffer[n] = '\0';
        for (j = 0; j < (sizeof(TRANSLATE_FUNC) / 5); j++) { // Check if Match in 3rd Translation table
          if (strncmp(buffer, TRANSLATE_FUNC[j], 4) == 0) { // Match?
            DigiKeyboard.sendKeyPress(TRANSLATE_FUNC[j][4], 0); // If found send KeyCode
            DigiKeyboard.sendKeyPress(0, 0);
            ESCFOUND = 0;
            n = 0;
            for (j = 0; j < 5; j++) {
              buffer[j] = '\0';
            }
          }
        }
      }
    } else if (c == 27) { // Detect ESC
      ESCFOUND = millis();
    } else {
      KeyboardWriteTranslate(c);
    }
    // mySerial.write(c); // Echo - don't do except for debug as the µC is busy with doing USB, too!
  }
  DigiKeyboard.delay(5);
}
